#include "playcap.h"

class CameraCapture::CGrabCB : public ISampleGrabberCB
{
public:
    STDMETHODIMP SampleCB(double SampleTime, IMediaSample *pSample)
    {
	return E_FAIL;
    }

    STDMETHODIMP BufferCB(double SampleTime, BYTE *pBuffer, long BufferLen)
    {
	_cc->new_frame(SampleTime, pBuffer, BufferLen);
	return S_OK;
    }

private:
    ULONG _ref_count;
    CameraCapture *_cc;

public:
    CGrabCB(CameraCapture *cc) {
	_cc = cc;
    }

    //! Return a ptr to a different COM interface to this object
    HRESULT APIENTRY QueryInterface( REFIID iid, void** ppvObject )
    {
	// Match the interface and return the proper pointer
	if ( iid == IID_IUnknown) {
	    *ppvObject = dynamic_cast<IUnknown*>( this );
	} else if ( iid == IID_ISampleGrabberCB ) {
	    *ppvObject = dynamic_cast<ISampleGrabberCB*>( this );
	} else {
	    // It didn't match an interface
	    *ppvObject = NULL;
	    return E_NOINTERFACE;
	}

	// Increment refcount on the pointer we're about to return
	this->AddRef();
	// Return success
	return S_OK;
    }

    ULONG APIENTRY AddRef() { return ++_ref_count; }
    ULONG APIENTRY Release() { return --_ref_count; }
};

const char *
CameraCapture::GetError()
{
    return _err.c_str();
}

bool
CameraCapture::FindCaptureDevices(std::vector<std::string> &retlist)
{
    USES_CONVERSION;

    HRESULT hr;
    ULONG cFetched;

    // Create the system device enumerator
    CComPtr <ICreateDevEnum> pDevEnum = NULL;

    hr = CoCreateInstance (CLSID_SystemDeviceEnum, NULL, CLSCTX_INPROC,
			   IID_ICreateDevEnum, (void ** ) &pDevEnum);
    if (FAILED(hr)) {
	char buf[10240];
	sprintf(buf, "Couldn't create system enumerator!  hr=0x%x", hr);
	_err = buf;
	return false;
    }

    // Create an enumerator for the video capture devices
    CComPtr <IEnumMoniker> pClassEnum = NULL;

    hr = pDevEnum->CreateClassEnumerator (CLSID_VideoInputDeviceCategory, &pClassEnum, 0);
    if (FAILED(hr)) {
	char buf[10240];
	sprintf(buf, "Couldn't create class enumerator!  hr=0x%x", hr);
	_err = buf;
	return false;
    }

    // If there are no enumerators for the requested type, then
    // CreateClassEnumerator will succeed, but pClassEnum will be NULL.
    if (pClassEnum == NULL) {
	_err =
	    "No video capture device was detected.\r\n\r\n"
	    "This sample requires a video capture device, such as a USB WebCam,\r\n"
	    "to be installed and working properly.  The sample will now close.";
	return false;
    }

    std::map<std::string, IBaseFilter *>::iterator it = _devices.begin();
    while (it != _devices.end()) { it->second->Release(); it++; }
    _devices.clear();

    // Use the first video capture device on the device list.
    // Note that if the Next() call succeeds but there are no monikers,
    // it will return S_FALSE (which is not a failure).  Therefore, we
    // check that the return code is S_OK instead of using SUCCEEDED() macro.
    while (1) {
	IBaseFilter * pSrc = NULL;
	CComPtr <IMoniker> pMoniker = NULL;
	if ((pClassEnum->Next (1, &pMoniker, &cFetched)) == S_OK) {
	    // Bind Moniker to a filter object
	    hr = pMoniker->BindToObject(0,0,IID_IBaseFilter, (void**)&pSrc);
	    if (FAILED(hr)) {
		char buf[10240];
		sprintf(buf, "Couldn't bind moniker to filter object!  hr=0x%x", hr);
		_err = buf;
		return false;
	    }

	    IPropertyBag *pBag;
	    hr = pMoniker->BindToStorage(0, 0, IID_IPropertyBag, (void **)&pBag);
	    if (SUCCEEDED(hr)) {
		VARIANT var;
		var.vt = VT_BSTR;
		hr = pBag->Read(L"FriendlyName", &var, NULL);
		retlist.push_back(W2T(var.bstrVal));
		_devices[W2T(var.bstrVal)] = pSrc;
	    } else {
		pSrc->Release();
	    }

	} else {
	    break;
	}
    }

    if (_devices.empty()) {
	_err =
	    "No video capture device was detected.\r\n\r\n"
	    "This sample requires a video capture device, such as a USB WebCam,\r\n"
	    "to be installed and working properly.  The sample will now close.";
	return false;
    }

    return true;
}

HRESULT
CameraCapture::GetInterfaces(void)
{
    HRESULT hr;

    // Create the filter graph
    hr = CoCreateInstance (CLSID_FilterGraph, NULL, CLSCTX_INPROC,
			   IID_IGraphBuilder, (void **) &_pGraph);
    if (FAILED(hr)) return hr;

    // Create the capture graph builder
    hr = CoCreateInstance (CLSID_CaptureGraphBuilder2 , NULL, CLSCTX_INPROC,
			   IID_ICaptureGraphBuilder2, (void **) &_pCapture);
    if (FAILED(hr)) return hr;

    // Obtain interfaces for media control and Video Window
    hr = _pGraph->QueryInterface(IID_IMediaControl,(LPVOID *) &_pMC);
    if (FAILED(hr)) return hr;

    return S_OK;
}


HRESULT
CameraCapture::GetPin(IBaseFilter *pFilter, PIN_DIRECTION PinDir, IPin **ppPin)
{
    IEnumPins  *pEnum;
    IPin       *pPin;
    pFilter->EnumPins(&pEnum);
    while (pEnum->Next(1, &pPin, 0) == S_OK)
    {
	PIN_DIRECTION PinDirThis;
	pPin->QueryDirection(&PinDirThis);
	if (PinDir == PinDirThis) {
	    pEnum->Release();
	    *ppPin = pPin;
	    return S_OK;
	}
	pPin->Release();
    }
    pEnum->Release();
    return E_FAIL;
}

HRESULT
CameraCapture::ConnectFilters(IGraphBuilder *pGraph,
			      IBaseFilter *pFirst, IBaseFilter *pSecond)
{
    IPin *pOut = NULL, *pIn = NULL;
    HRESULT hr = GetPin(pFirst, PINDIR_OUTPUT, &pOut);
    if (FAILED(hr)) return hr;
    hr = GetPin(pSecond, PINDIR_INPUT, &pIn);
    if (FAILED(hr))
    {
	pOut->Release();
	return E_FAIL;
    }
    hr = pGraph->Connect(pOut, pIn);
    pIn->Release();
    pOut->Release();
    return hr;
}

HRESULT
CameraCapture::AddGraphToRot(IUnknown *pUnkGraph, DWORD *pdwRegister)
{
    IMoniker * pMoniker;
    IRunningObjectTable *pROT;
    WCHAR wsz[128];
    HRESULT hr;

    if (FAILED(GetRunningObjectTable(0, &pROT))) {
	return E_FAIL;
    }

    wsprintfW(wsz, L"FilterGraph %08x pid %08x", (DWORD_PTR)pUnkGraph,
	      GetCurrentProcessId());

    hr = CreateItemMoniker(L"!", wsz, &pMoniker);
    if (SUCCEEDED(hr)) {
	hr = pROT->Register(0, pUnkGraph, pMoniker, pdwRegister);
	pMoniker->Release();
    }
    pROT->Release();
    return hr;
}

void
CameraCapture::RemoveGraphFromRot(DWORD pdwRegister)
{
    IRunningObjectTable *pROT;
    if (SUCCEEDED(GetRunningObjectTable(0, &pROT))) {
	pROT->Revoke(pdwRegister);
	pROT->Release();
    }
}

CameraCapture::CameraCapture() {
    _mygrab = new CGrabCB(this);
    _dwGraphRegister = 0;
    _pMC = NULL;
    _pGraph = NULL;
    _pCapture = NULL;
    _pGrabber = NULL;
}

CameraCapture::~CameraCapture() {
    std::map<std::string, IBaseFilter *>::iterator it = _devices.begin();
    while (it != _devices.end()) { it->second->Release(); it++; }
    _devices.clear();

    if (_pGrabber) _pGrabber->SetCallback(NULL, 1);
    delete _mygrab;

    // Stop previewing data
    if (_pMC) _pMC->StopWhenReady();

    // Remove filter graph from the running object table
    if (_dwGraphRegister) RemoveGraphFromRot(_dwGraphRegister);

    // Release DirectShow interfaces
    if (_pGrabber) _pGrabber->Release();
    if (_pMC) _pMC->Release();
    if (_pGraph) _pGraph->Release();
    if (_pCapture) _pCapture->Release();
}

bool CameraCapture::Init(const char *devname) {
    HRESULT hr;
    IBaseFilter *pSrcFilter = NULL;

    // Get DirectShow interfaces
    hr = GetInterfaces();
    if (FAILED(hr)) {
	char buf[10240];
	sprintf(buf, "Failed to get video interfaces!  hr=0x%x", hr);
	_err = buf;
	return false;
    }

    // Attach the filter graph to the capture graph
    hr = _pCapture->SetFiltergraph(_pGraph);
    if (FAILED(hr)) {
	char buf[10240];
	sprintf(buf, "Failed to set capture filter graph!  hr=0x%x", hr);
	_err = buf;
	return false;
    }

    // Use the system device enumerator and class enumerator to find
    // a video capture/preview device, such as a desktop USB video camera.
    if (_devices.find(devname) == _devices.end()) {
	char buf[10240];
	sprintf(buf, "Couldn't find capture device: \"%s\"", devname);
	_err = buf;
	return false;
    }
    pSrcFilter = _devices[devname];
    if (pSrcFilter == NULL) {
	// Don't display a message because FindCaptureDevice will handle it
	return false;
    }

    // Add Capture filter to our graph.
    hr = _pGraph->AddFilter(pSrcFilter, L"Video Capture");
    if (FAILED(hr)) {
	char buf[10240];
	sprintf(buf, "Couldn't add capture filter to graph!  hr=0x%x", hr);
	_err = buf;
	return false;
    }

    {
	IBaseFilter *pF = NULL;
	hr = CoCreateInstance(CLSID_SampleGrabber, NULL, CLSCTX_INPROC_SERVER,
			      IID_IBaseFilter, reinterpret_cast<void**>(&pF));
	hr = pF->QueryInterface(IID_ISampleGrabber,
				reinterpret_cast<void**>(&_pGrabber));

	AM_MEDIA_TYPE mt;
	ZeroMemory(&mt, sizeof(AM_MEDIA_TYPE));
	mt.majortype = MEDIATYPE_Video;
	mt.subtype = MEDIASUBTYPE_RGB24;
	mt.formattype = FORMAT_VideoInfo;
	_pGrabber->SetMediaType(&mt);
	_pGrabber->SetCallback(_mygrab, 1);

	hr = _pGraph->AddFilter(pF, L"SampleGrabber");
	if (FAILED(hr)) {
	    char buf[10240];
	    sprintf(buf, "Couldn't add hook filter to graph!  hr=0x%x", hr);
	    _err = buf;
	    return false;
	}

	ConnectFilters(_pGraph, pSrcFilter, pF);

	ZeroMemory(&mt, sizeof(AM_MEDIA_TYPE));
	hr = _pGrabber->GetConnectedMediaType(&mt);
	VIDEOINFOHEADER *pVih = (VIDEOINFOHEADER*)(mt.pbFormat);
	_width = pVih->bmiHeader.biWidth;
	_height = pVih->bmiHeader.biHeight;

	// We want the absolute height, and don't care about up-down orientation.
	//	if (height < 0) height *= -1;
	//	FreeMediaType(mt);
//	printf("%d,%d\n", _width, _height);
//	fflush(stdout);

	IBaseFilter *pNull = NULL;
	hr = CoCreateInstance(CLSID_NullRenderer, NULL, CLSCTX_INPROC_SERVER,
			      IID_IBaseFilter, reinterpret_cast<void**>(&pNull));
	hr = _pGraph->AddFilter(pNull, L"NullRenderer");
	hr = ConnectFilters(_pGraph, pF, pNull);
	pNull->Release();
	pF->Release();
    }

    // Add our graph to the running object table, which will allow
    // the GraphEdit application to "spy" on our graph
    hr = AddGraphToRot(_pGraph, &_dwGraphRegister);
    if (FAILED(hr)) {
	_dwGraphRegister = 0;
	char buf[10240];
	sprintf(buf, "Failed to register filter graph with ROT!  hr=0x%x", hr);
	_err = buf;
	return false;
    }

    return true;
}

bool CameraCapture::Start(void) {
    HRESULT hr = _pMC->Run();
    if (FAILED(hr)) {
	char buf[10240];
	sprintf(buf, "Couldn't run the graph!  hr=0x%x", hr);
	_err = buf;
	return false;
    }

    return true;
}

bool CameraCapture::Stop(void) {
    HRESULT hr = _pMC->StopWhenReady();
    if (FAILED(hr)) {
	char buf[10240];
	sprintf(buf, "Couldn't stop the graph!  hr=0x%x", hr);
	_err = buf;
	return false;
    }

    return true;
}

class mycap : public CameraCapture {
public:
    void new_frame(double SampleTime, BYTE *buffer, long len) {
	printf("%d, ", len);
	fflush(stdout);
    }
};

/*
int APIENTRY
WinMain(HINSTANCE hInstance, HINSTANCE hInstP, LPSTR lpCmdLine, int nCmdShow)
{
    CoInitialize(NULL);

    mycap c;

    if (!c.Init())
	exit(1);

    c.Start();

    MSG msg;
    while(GetMessage(&msg,NULL,0,0)) {
	TranslateMessage(&msg);
	DispatchMessage(&msg);
    }

    CoUninitialize();
    return 0;
}
 */
